//
//  GTReference.m
//  ObjectiveGitFramework
//
//  Created by Timothy Clem on 3/2/11.
//
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
//

#import "GTReference.h"
#import "GTOID.h"
#import "GTReflog+Private.h"
#import "GTRepository.h"
#import "NSError+Git.h"
#import "NSString+Git.h"

@implementation GTReference

- (NSString *)description {
  return [NSString stringWithFormat:@"<%@: %p> type: %@, repository: %@", NSStringFromClass([self class]), self, self.type, self.repository];
}

- (void)dealloc {
	if(self.git_reference != NULL) {
		git_reference_free(self.git_reference);
		_git_reference = NULL;
	}
}


#pragma mark API

- (BOOL)isRemote {
	return git_reference_is_remote(self.git_reference) != 0;
}

+ (id)referenceByLookingUpReferencedNamed:(NSString *)refName inRepository:(GTRepository *)theRepo error:(NSError **)error {
	return [[self alloc] initByLookingUpReferenceNamed:refName inRepository:theRepo error:error];
}

+ (id)referenceByCreatingReferenceNamed:(NSString *)refName fromReferenceTarget:(NSString *)target inRepository:(GTRepository *)theRepo error:(NSError **)error {
	return [[self alloc] initByCreatingReferenceNamed:refName fromReferenceTarget:target inRepository:theRepo error:error];
}

+ (id)referenceByResolvingSymbolicReference:(GTReference *)symbolicRef error:(NSError **)error {	
	return [[self alloc] initByResolvingSymbolicReference:symbolicRef error:error];
}

- (id)initByLookingUpReferenceNamed:(NSString *)refName inRepository:(GTRepository *)repo error:(NSError **)error {
	NSParameterAssert(refName != nil);
	NSParameterAssert(repo != nil);

	git_reference *ref = NULL;
	int gitError = git_reference_lookup(&ref, repo.git_repository, refName.UTF8String);
	if (gitError != GIT_OK) {
		if (error != NULL) *error = [NSError git_errorFor:gitError withAdditionalDescription:@"Failed to lookup reference."];
		return nil;
	}

	return [self initWithGitReference:ref repository:repo];
}

- (id)initByCreatingReferenceNamed:(NSString *)refName fromReferenceTarget:(NSString *)target inRepository:(GTRepository *)repo error:(NSError **)error {
	NSParameterAssert(refName != nil);
	NSParameterAssert(target != nil);
	NSParameterAssert(repo != nil);

	git_oid oid;
	int gitError = git_oid_fromstr(&oid, target.UTF8String);
	git_reference *ref = NULL;
	if (gitError == GIT_OK) {
		gitError = git_reference_create(&ref, repo.git_repository, refName.UTF8String, &oid, 0);
	} else {
		gitError = git_reference_symbolic_create(&ref, repo.git_repository, refName.UTF8String, target.UTF8String, 0);
	}

	if (gitError != GIT_OK) {
		if(error != NULL) *error = [NSError git_errorFor:gitError withAdditionalDescription:@"Failed to create symbolic reference."];
		return nil;
	}

	return [self initWithGitReference:ref repository:repo];
}

- (id)initByResolvingSymbolicReference:(GTReference *)symbolicRef error:(NSError **)error {
	NSParameterAssert(symbolicRef != nil);

	git_reference *ref = NULL;
	int gitError = git_reference_resolve(&ref, symbolicRef.git_reference);
	if (gitError != GIT_OK) {
		if (error != NULL) *error = [NSError git_errorFor:gitError withAdditionalDescription:@"Failed to resolve reference."];
		return nil;
	}

	return [self initWithGitReference:ref repository:symbolicRef.repository];
}

- (id)initWithGitReference:(git_reference *)ref repository:(GTRepository *)repo {
	NSParameterAssert(ref != NULL);
	NSParameterAssert(repo != nil);

	self = [super init];
	if (self == nil) return nil;

	_git_reference = ref;
	_repository = repo;

	return self;
}

- (NSString *)name {
	const char *refName = git_reference_name(self.git_reference);
	if (refName == NULL) return nil;
	
	return @(refName);
}

- (GTReference *)referenceByRenaming:(NSString *)newName error:(NSError **)error {
	NSParameterAssert(newName != nil);
	
	git_reference *newRef = NULL;
	int gitError = git_reference_rename(&newRef, self.git_reference, newName.UTF8String, 0);
	if (gitError != GIT_OK) {
		if (error != NULL) *error = [NSError git_errorFor:gitError withAdditionalDescription:@"Failed to rename reference."];
		return NO;
	}

	return [[self.class alloc] initWithGitReference:newRef repository:self.repository];
}

- (NSString *)type {
	return @(git_object_type2string((git_otype)git_reference_type(self.git_reference)));
}

- (NSString *)target {
	if (git_reference_type(self.git_reference) == GIT_REF_OID) {
		return [NSString git_stringWithOid:git_reference_target(self.git_reference)];
	} else {
		return @(git_reference_symbolic_target(self.git_reference));
	}
}

- (GTReference *)referenceByUpdatingTarget:(NSString *)newTarget error:(NSError **)error {
	NSParameterAssert(newTarget != nil);

	int gitError;
	git_reference *newRef = NULL;
	if (git_reference_type(self.git_reference) == GIT_REF_OID) {
		git_oid oid;
		gitError = git_oid_fromstr(&oid, newTarget.UTF8String);
		if (gitError != GIT_OK) {
			if(error != NULL) *error = [NSError git_errorFor:gitError withAdditionalDescription:[NSString stringWithFormat:@"Failed to create OID from string: %@", newTarget]];
			return nil;
		}
		
		gitError = git_reference_set_target(&newRef, self.git_reference, &oid);
	} else {
		gitError = git_reference_symbolic_set_target(&newRef, self.git_reference, newTarget.UTF8String);
	}

	if (gitError != GIT_OK) {
		if (error != NULL) *error = [NSError git_errorFor:gitError withAdditionalDescription:@"Failed to set reference target."];
		return NO;
	}

	return [[self.class alloc] initWithGitReference:newRef repository:self.repository];
}

- (BOOL)deleteWithError:(NSError **)error {
	int gitError = git_reference_delete(self.git_reference);
	if (gitError != GIT_OK) {
		if (error != NULL) *error = [NSError git_errorFor:gitError withAdditionalDescription:@"Failed to delete reference."];
		return NO;
	}

	return YES;
}

- (GTReference *)resolvedReferenceWithError:(NSError **)error {
	return [GTReference referenceByResolvingSymbolicReference:self error:error];
}

- (const git_oid *)git_oid {
	return git_reference_target(self.git_reference);
}

- (GTOID *)OID {
	const git_oid *oid = self.git_oid;
	if (oid == NULL) return nil;

	return [[GTOID alloc] initWithGitOid:oid];
}

- (GTReference *)reloadedReferenceWithError:(NSError **)error {
	return [[self.class alloc] initByLookingUpReferenceNamed:self.name inRepository:self.repository error:error];
}

+ (NSError *)invalidReferenceError {
	return [NSError git_errorFor:GTReferenceErrorCodeInvalidReference withAdditionalDescription:@"Invalid git_reference."];
}

- (GTReflog *)reflog {
	return [[GTReflog alloc] initWithReference:self];
}

@end
